package data.missions.aiw_spiretest;

import com.fs.starfarer.api.campaign.CargoAPI.CrewXPLevel;
import com.fs.starfarer.api.fleet.FleetGoal;
import com.fs.starfarer.api.fleet.FleetMemberAPI;
import com.fs.starfarer.api.fleet.FleetMemberType;
import com.fs.starfarer.api.mission.FleetSide;
import com.fs.starfarer.api.mission.MissionDefinitionAPI;
import com.fs.starfarer.api.mission.MissionDefinitionPlugin;
import java.util.ArrayList;
import java.util.List;

public class MissionDefinition implements MissionDefinitionPlugin {

	private List ships = new ArrayList();
	private void addShip(String variant, int weight) {
		for (int i = 0; i < weight; i++) {
			ships.add(variant);
		}
	}
	
	private void generateFleet(int maxFP, FleetSide side, List ships, MissionDefinitionAPI api) {
		int currFP = 0;
		while (true) {
			int index = (int)(Math.random() * ships.size());
			String id = (String) ships.get(index);
			currFP += api.getFleetPointCost(id);
			if (currFP > maxFP) {
				return;
			}
			
			if (id.endsWith("_wing")) {
				api.addToFleet(side, id, FleetMemberType.FIGHTER_WING, false);
			} else {
				api.addToFleet(side, id, FleetMemberType.SHIP, false);
			}
		}
	}
	
	public void defineMission(MissionDefinitionAPI api) 
	{
		// defs for random ships
		addShip("doom_Strike", 3);
		addShip("shade_Assault", 7);
		addShip("afflictor_Strike", 7);
		addShip("hyperion_Attack", 3);
		addShip("hyperion_Strike", 3);
		addShip("onslaught_Standard", 3);
		addShip("onslaught_Outdated", 3);
		addShip("onslaught_Elite", 1);
		addShip("astral_Elite", 3);
		addShip("paragon_Elite", 1);
		addShip("odyssey_Balanced", 2);
		addShip("conquest_Elite", 3);
		addShip("eagle_Assault", 5);
		addShip("falcon_Attack", 5);
		addShip("venture_Balanced", 5);
		addShip("apogee_Balanced", 5);
		addShip("aurora_Balanced", 5);
		addShip("aurora_Balanced", 5);
		addShip("dominator_Assault", 5);
		addShip("dominator_Support", 5);
		addShip("medusa_Attack", 5);
		addShip("condor_Strike", 15);
		addShip("condor_FS", 15);
		addShip("enforcer_Assault", 15);
		addShip("enforcer_CS", 15);
		addShip("hammerhead_Balanced", 10);
		addShip("hammerhead_Elite", 5);
		addShip("sunder_CS", 10);
		addShip("gemini_Standard", 8);
		addShip("buffalo2_FS", 20);
		addShip("lasher_CS", 20);
		addShip("lasher_Standard", 20);
		addShip("hound_Standard", 15);
		addShip("tempest_Attack", 15);
		addShip("brawler_Assault", 15);
		addShip("wolf_CS", 2);
		addShip("hyperion_Strike", 1);
		addShip("vigilance_Standard", 10);
		addShip("vigilance_FS", 15);
		addShip("tempest_Attack", 2);
		addShip("brawler_Assault", 10);
		addShip("piranha_wing", 15);
		addShip("talon_wing", 20);
		addShip("broadsword_wing", 10);
		addShip("mining_drone_wing", 10);
		addShip("wasp_wing", 10);
		addShip("xyphos_wing", 10);
		addShip("longbow_wing", 10);
		addShip("dagger_wing", 10);
		addShip("thunder_wing", 5);
		addShip("gladius_wing", 15);
		addShip("warthog_wing", 5);

		// Set up the fleets so we can add ships and fighter wings to them.
		api.initFleet(FleetSide.PLAYER, "SIS", FleetGoal.ATTACK, false, 5);
		api.initFleet(FleetSide.ENEMY, "ISS", FleetGoal.ATTACK, true, 5);

		// Set a small blurb for each fleet that shows up on the mission detail and
		// mission results screens to identify each side.
		api.setFleetTagline(FleetSide.PLAYER, "Spire forces");
		api.setFleetTagline(FleetSide.ENEMY, "Opposing force");
		
		// These show up as items in the bulleted list under 
		// "Tactical Objectives" on the mission detail screen
		api.addBriefingItem("Defeat all enemy forces");
		
		// Set up the player's fleet.  Variant names come from the
		// files in data/variants and data/variants/fighters
		api.addToFleet(FleetSide.PLAYER, "spire_battleship_Command", FleetMemberType.SHIP, true);
		api.addToFleet(FleetSide.PLAYER, "spire_champion_Elite", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.PLAYER, "spire_cruiser_Line", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.PLAYER, "spire_destroyer_Assault", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.PLAYER, "spire_destroyer_Support", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.PLAYER, "spire_lcarrier_Standard", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.PLAYER, "spire_frigate_Standard", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.PLAYER, "spire_frigate_Skirmisher", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.PLAYER, "spire_corvette_Standoff", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.PLAYER, "spire_corvette_Lance", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.PLAYER, "spire_armorrot_wing", FleetMemberType.FIGHTER_WING, false);
		api.addToFleet(FleetSide.PLAYER, "spire_armorrot_wing", FleetMemberType.FIGHTER_WING, false);
		api.addToFleet(FleetSide.PLAYER, "spire_colonyship_Standard", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.PLAYER, "spire_colonyship_Hammer", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.PLAYER, "spire_stealthship_Strike", FleetMemberType.SHIP, false);

		// Set up the enemy fleet.
		/*
		api.addToFleet(FleetSide.ENEMY, "onslaught_Standard", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.ENEMY, "aurora_Balanced", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.ENEMY, "heron_Standard", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.ENEMY, "hammerhead_Elite", FleetMemberType.SHIP, false);
		//api.addToFleet(FleetSide.ENEMY, "sunder_Assault", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.ENEMY, "enforcer_CS", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.ENEMY, "wolf_Assault", FleetMemberType.SHIP, false);
		//api.addToFleet(FleetSide.ENEMY, "omen_PD", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.ENEMY, "vigilance_FS", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.ENEMY, "lasher_Strike", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.ENEMY, "hound_Standard", FleetMemberType.SHIP, false);
		api.addToFleet(FleetSide.ENEMY, "thunder_wing", FleetMemberType.FIGHTER_WING, false);
		api.addToFleet(FleetSide.ENEMY, "gladius_wing", FleetMemberType.FIGHTER_WING, false);
		api.addToFleet(FleetSide.ENEMY, "dagger_wing", FleetMemberType.FIGHTER_WING, false);
		*/
		generateFleet(150 + (int)((float) Math.random() * 100), FleetSide.ENEMY, ships, api);
		
		// Set up the map.
		float width = 16000f;
		float height = 16000f;
		api.initMap((float)-width/2f, (float)width/2f, (float)-height/2f, (float)height/2f);
		
		float minX = -width/2;
		float minY = -height/2;
		
		// All the addXXX methods take a pair of coordinates followed by data for
		// whatever object is being added.
		
		// Add two big nebula clouds
		api.addNebula(minX + width * 0.75f, minY + height * 0.7f, 1600);
		api.addNebula(minX + width * 0.25f, minY + height * 0.4f, 1200);
		
		// And a few random ones to spice up the playing field.
		// A similar approach can be used to randomize everything
		// else, including fleet composition.
		for (int i = 0; i < 7; i++) {
			float x = (float) Math.random() * width - width/2;
			float y = (float) Math.random() * height - height/2;
			float radius = 200f + (float) Math.random() * 600f; 
			api.addNebula(x, y, radius);
		}
		
		// Add objectives
		api.addObjective(minX + width * 0.35f, minY + height * 0.4f, "sensor_array");
		api.addObjective(minX + width * 0.7f, minY + height * 0.6f, "nav_buoy");
		
		// Add asteroid field
		//api.addAsteroidField(minX, minY + height * 0.5f, 0, height, 20f, 70f, 50);
		
		// Add planet
		String [] planets = {"barren", "terran", "gas_giant", "ice_giant", "cryovolcanic", "frozen", "jungle", "desert", "arid"};
		String planet = planets[(int) (Math.random() * (double) planets.length)];
		float radius = 100f + (float) Math.random() * 150f;
		api.addPlanet(0, 0, radius, planet, 200f, true);
	}

}
